// Code is written by Jianing Wei
// Unauthorized use or copying is prohibited
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "../../src/allocate.h"
#include "../../src/SpaceVaryingConv.h"
#include "../../src/psf.h"
#include "../../src/FastResize.h"

#define NUMEVALUATION 1000
#define MAXVAL 2000000
#define SAMPLES 1048567
#define IMG_HEIGHT 4.31
#define NUMOFSTAGES 2


void Usage(char *argv[])
{
  printf("Wrong number of arguments. Usage:\n");
  printf("%s img.pgm listmatrix.dat scaling_factors.txt [nonzero_index.txt]\n", argv[0]);
  printf("Try again ...\n");
  exit(-1);
}


unsigned char * readPGM(char *pgmfile, int *pheight, int *pwidth)
{
  FILE *fp;
  char pgm_symbol[3];
  int maxval;
  unsigned char *pdata;
  int num_read;

  if ( (fp = fopen(pgmfile, "r")) ) {
    fscanf(fp, "%s", pgm_symbol);
    fscanf(fp, "%d %d", pwidth, pheight);
    fscanf(fp, "%d\n", &maxval);
    pdata = (unsigned char *) malloc(sizeof(unsigned char)*(*pheight)*(*pwidth));
    num_read = fread(pdata, sizeof(unsigned char), (*pheight)*(*pwidth), 
	             fp);
    fclose(fp);
    if (num_read!=(*pheight)*(*pwidth)) {
      printf("Image read failed\n");
      exit(-1);
    }
  }
  else {
    printf("cannot open file %s to read\n", pgmfile);
    exit(-1);
  }

  return pdata;
}


void GenerateRand2dIndex(unsigned int *index2d, int height, int width, int num_index)
{
  int i, num_of_bits=0, bitshift;
  for (num_of_bits=0; (RAND_MAX>>num_of_bits)>0; num_of_bits++);

  bitshift = num_of_bits+16-sizeof(int)*8;
  bitshift = bitshift>0?bitshift:0;
  for (i=0; i<num_index; i++) {
    index2d[2*i] = (rand()>>bitshift)*(height-1)/(RAND_MAX>>bitshift);
    index2d[2*i+1] = (rand()>>bitshift)*(width-1)/(RAND_MAX>>bitshift);
  }  
}


static float ComputeNRMSE(float **approx_output, float *exact_output, 
                          unsigned int *pixel_index, int num_pixels)
{
  int i, j, n;
  float energy, sqdiff, nrmse;
  energy = 0;
  sqdiff = 0;
  for ( n=0; n<num_pixels; n=n+1 ) {
    i = pixel_index[2*n];
    j = pixel_index[2*n+1];
    energy += exact_output[n]*exact_output[n];
    sqdiff += (exact_output[n]-approx_output[i][j])*(exact_output[n]-approx_output[i][j]);
  }
  nrmse = sqrt(sqdiff/energy);
  return nrmse;
}


int main(int argc, char *argv[])
{
  int i, j;
  int height, width, height_matched, width_matched;
  struct parameter para;
  float delta;
  int num_levels, block_num_levels, block_size, height_padded, width_padded;
  unsigned char *pimg;
  float **input_img, **output_img, **matched_img, **output_org;
  int num_evaluation;
  unsigned int *pixel_index;
  float *pixel_val;
  LIST **listmatrix;
  float *scaling_factors;
  float nrmse;
  char *pchar;
  char input_img_file[128], listmatrix_file[128], scaling_factors_file[128];
  char debug_out_index_file[128], ground_truth_file[128];
  FILE *fp;
  struct power_LUT *powertable;
  float pixel_size;
  cimage_uchar input_uchar, input_temp;
  cimage out_temp_small, out_temp_full;

  if (argc<4 || argc>5) {
    Usage(argv);
  }
  strcpy(input_img_file, argv[1]);
  strcpy(listmatrix_file, argv[2]);
  strcpy(scaling_factors_file, argv[3]);
  if (argc==5) {
	strcpy(debug_out_index_file, argv[4]);
  }
  strcpy(ground_truth_file, input_img_file);
  pchar = strrchr(ground_truth_file, '.');
  if (pchar!=NULL) {
	*pchar = '\0';
  }
  strcat(ground_truth_file, "_ground_truth.txt");

  if ( !(pchar = strstr(input_img_file, ".pgm")) ) {
    printf("Wrong file type!\n");
    exit(-1);
  }

  
  
  if ( (fp=fopen(listmatrix_file, "r")) ) {
    if ( ReadListmatrixHeader( fp, &para, &delta, &num_levels, 
                               &block_num_levels, &height_padded, &width_padded, &block_size ) ) {
      printf("Read PSF parameters, a=%e, b=%e, f=%f, g=%f\n", para.a, para.b, para.f, para.g);
      listmatrix = (LIST**) multialloc(sizeof(LIST), 2, height_padded, width_padded);
      for ( i=0; i<height_padded; i++) {
        for ( j=0; j<width_padded; j++) {
          listmatrix[i][j].head = NULL;
          listmatrix[i][j].tail = NULL;
          listmatrix[i][j].num_pix = 0;
        }
      }
      ReadListmatrix( fp, listmatrix, height_padded, width_padded );
    }
    else {
      printf("Cannot successfully read stored matrix\n");
      exit(-1);
    }
    fclose(fp);
  }
  else {
    printf("cannot open file %s to read\n", listmatrix_file);
    exit(-1);
  }

  // obtain input image, resize if necessary
  pimg =  readPGM( input_img_file, &height, &width );
  input_img = multialloc(sizeof(float), 2, height, width);
  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      input_img[i][j] = (float)pimg[i*width+j];
    }
  }
  if ( height>height_padded ) {
    input_uchar.ptr = pimg;
    input_uchar.height = height;
    input_uchar.width = width;
    initUcharImage(&input_temp, height_padded, width_padded);
    fast_resize(&input_uchar, &input_temp);
    matched_img = multialloc(sizeof(float), 2, input_temp.height, input_temp.width);
    for (i=0; i<input_temp.height; i++) {
      for (j=0; j<input_temp.width; j++) {
    	matched_img[i][j] = (float)imaccess(&input_temp, j, i);
      }
    }
    destroyUcharImage(&input_temp);
    height_matched = height_padded;
    width_matched = width_padded;
  }
  else {
    matched_img = multialloc(sizeof(float), 2, height, width);
    for (i=0; i<height; i++) {
      for (j=0; j<width; j++) {
        matched_img[i][j] = input_img[i][j];
      }
    }
    height_matched = height;
    width_matched = width;
  }

  

  scaling_factors = (float*) malloc(sizeof(float)*(1+3*num_levels));
  if ( (fp=fopen(scaling_factors_file, "r")) ) {
    for (i=0; i<1+3*num_levels; i++) {
      fscanf(fp, "%f", scaling_factors+i);
    }
    fclose(fp);
  }
  else {
    printf("cannot open file %s to read\n", scaling_factors_file);
    exit(-1);
  }

/*
  fp = fopen("input.txt", "w");
  for (i=0; i<height_matched; i++) {
    for (j=0; j<width_matched; j++) {
      fprintf(fp, "%f ", matched_img[i][j]);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);
*/

  output_img = multialloc(sizeof(float), 2, height_matched, width_matched);
  SpaceVaryingTransform( matched_img, height_matched, width_matched, listmatrix, 
                         height_padded, width_padded, block_size, block_num_levels,   
                         scaling_factors, num_levels, output_img, NUMOFSTAGES );

  fp = fopen("output.txt", "w");
  for (i=0; i<height_matched; i++) {
    for (j=0; j<width_matched; j++) {
      fprintf(fp, "%f ", output_img[i][j]);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);


  if (argc<5) {
    num_evaluation = NUMEVALUATION;
    pixel_index = (unsigned int *)malloc(sizeof(unsigned int)*num_evaluation*2);
    pixel_val = (float *)malloc(sizeof(float)*num_evaluation);
    srand(1);
    GenerateRand2dIndex(pixel_index, height, width, num_evaluation);
 
    /* 
    sprintf(debug_out_index_file, "imgnonzero_%dx%d.txt", height, width);
    if ( ( fp = fopen(debug_out_index_file, "w")) ) {
      for (i=0; i<NUMEVALUATION; i++) {
        fprintf(fp, "%u %u\n", *(pixel_index+2*i), *(pixel_index+2*i+1));
      }
      fclose(fp);
    }
    else {
      printf("cannot open file to write\n");
      exit(-1);
    }
    */
  }
  else {
    if ( (fp = fopen(debug_out_index_file, "r")) ) {
      fscanf(fp, "%d\n", &num_evaluation);
      pixel_index = (unsigned int *)malloc(sizeof(unsigned int)*num_evaluation*2);
      pixel_val = (float *)malloc(sizeof(float)*num_evaluation);
      for (i=0; i<num_evaluation; i++) {
        fscanf(fp, "%u %u\n", pixel_index+2*i, pixel_index+2*i+1);
      }
      fclose(fp);
    }
    else {
	  printf("cannot open file %s to read\n", debug_out_index_file);
	  exit(-1);
    }
  }
  

  powertable = build_power_LUT(para.f, MAXVAL, SAMPLES);
  pixel_size = IMG_HEIGHT/height;
  RigoConv( para, powertable, height, width, input_img, pixel_val, 
            pixel_index, num_evaluation, pixel_size);
  if ( fp=fopen(ground_truth_file, "w") ) {
    for (i=0; i<num_evaluation; i++) {
      fprintf(fp, "%f\n", pixel_val[i]);
    }
    fclose(fp);  
  }
  else {
	printf("cannot open file %s to write\n", ground_truth_file);
	exit(-1);
  }

  multifree(input_img, 2);
  multifree(matched_img, 2);

  output_org = multialloc(sizeof(float), 2, height, width);
  if (height>height_padded) {
	initImage(&out_temp_small, height_matched, width_matched);
	for (i=0; i<height_matched; i++) {
	  for (j=0; j<width_matched; j++) {
		imaccess(&out_temp_small, j, i) = output_img[i][j];
	  }
	}
	initImage(&out_temp_full, height, width);
	fast_resize_float(&out_temp_small, &out_temp_full);
	destroyImage(&out_temp_small);
	for (i=0; i<height; i++) {
	  for (j=0; j<width; j++) {
		output_org[i][j] = imaccess(&out_temp_full, j, i);
	  }
	}
	destroyImage(&out_temp_full);
  }
  else {
	for (i=0; i<height; i++) {
	  for (j=0; j<width; j++) {
		output_org[i][j] = output_img[i][j];
	  }
	}
  }

  nrmse = ComputeNRMSE( output_org, pixel_val, pixel_index, num_evaluation );
  printf("NRMSE = %7.5f\n", nrmse);
  
/*
  fp = fopen("test.txt", "w");
  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      fprintf(fp, "%8.4f ", output_org[i][j]);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);
*/  

  free(pimg);
  multifree(output_org, 2);
  multifree(output_img, 2);
  
  multifree(listmatrix, 2);
  free(scaling_factors);
  distroy_power_LUT(powertable);
  free(pixel_index);
  free(pixel_val);

  return 1;
}
