/* Jianing Wei Nov 21, 2008 */
/* This code implements regular Haar wavelet transform with zero boundary condition */
/* Before compiling and executing, check PSF_IMG. If it is 1, then transform 
   PSF of given parameters; if it is 0, transform the input image. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <math.h>
#include "power.h"

#define MAXVAL 2000000 /* max value used in power function */
#define SAMPLES 1048567 /* quantization levels in power LUT */
#define IMG_height 4.31
#define PSF_IMG 1


struct parameter {
  float a;
  float b;
  float f;
  float g;
};

typedef struct parameter PARAMETER;


float computepsf(PARAMETER para, struct power_LUT *powertable, 
		    float x0, float y0, float x, float y, float pixel_size)
{
  float r2, term1, term2, value, norm_constant;
  float temp[1];
  x = x*pixel_size;
  y = y*pixel_size;
  x0 = x0*pixel_size;
  y0 = y0*pixel_size;

  r2 = x0*x0+y0*y0;
 
  norm_constant = 3.1415926*(para.g+para.a*r2)*(para.g+para.b*r2)/(para.f-1);
  norm_constant = norm_constant/pixel_size/pixel_size;
 
  if (r2==0) {
    temp[0] = 1+(x*x+y*y)/(para.g*para.g);
    if (temp[0]>MAXVAL) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  else {
    term1 = (x*x0+y*y0-r2)/(para.g+para.a*r2);
    term2 = (-x*y0+y*x0)/(para.g+para.b*r2);
    temp[0] = 1+(term1*term1+term2*term2)/r2;
    if (temp[0]>MAXVAL) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  return value;

}

void conv_noext(float *convdata, float *data, float *filter,
                int filterlen, int filter_origin, int width)
{
  int i, j;
  for ( i=0; i<width; i=i+1 ) {
    for ( j=0; j<filterlen; j=j+1 ) {
      if ( i+j-filter_origin>=0 && i+j-filter_origin<width ) {
        convdata[i+j-filter_origin] = convdata[i+j-filter_origin] + data[i]*filter[j];
      }
    }
  }
}

/* symmetric extention */
/*
void ext_conv_downsp(float *data, int length, float *filter, int filter_len, int start, 
                     float *conved)
{
  int i, j;
  float temp;
  int extended_len = length + 2*(filter_len-1);
  float *extended_data;

  extended_data = malloc(sizeof(float)*extended_len);
  for ( i=filter_len-1; i<length+filter_len-1; i=i+1 ) {
    extended_data[i] = data[i-filter_len+1];
  }
  for ( i=0; i<filter_len-1; i=i+1 ) {
    extended_data[filter_len-2-i] = data[i];
    extended_data[filter_len-1+length+i] = data[length-1-i];
  }
  for ( i=filter_len-1+start; i<extended_len; i=i+2 ) {
    temp = 0;
    for ( j=0; j<filter_len; j=j+1 ) {
      temp = temp + extended_data[i+j-filter_len+1]*filter[j];
    }
    conved[(i-filter_len+1-start)/2] = temp;
  }
  free(extended_data);
  
}
*/

/* zero extention */
void ext_conv_downsp(float *data, int length, float *filter, int filter_len, int start,
                     float *conved)
{
  int i, j;
  float temp;
  int extended_len = length + 2*(filter_len-1);
  float *extended_data;
                                                                                                                       
  extended_data = malloc(sizeof(float)*extended_len);
  for ( i=filter_len-1; i<length+filter_len-1; i=i+1 ) {
    extended_data[i] = data[i-filter_len+1];
  }
  for ( i=0; i<filter_len-1; i=i+1 ) {
    extended_data[filter_len-2-i] = 0;
    extended_data[filter_len-1+length+i] = 0;
  }
  for ( i=filter_len-1+start; i<extended_len; i=i+2 ) {
    temp = 0;
    for ( j=0; j<filter_len; j=j+1 ) {
      temp = temp + extended_data[i+j-filter_len+1]*filter[j];
    }
    conved[(i-filter_len+1-start)/2] = temp;
  }
  free(extended_data);
                                                                                                                       
}


/* this haar transform is supposed to handle all kinds of image sizes */
void new_haar_transform(float **data, int levels, int start, 
                        int **band_size, int **band_pos, float **wave_coef)
{
  int i, j, n;
  float **l, **h, **tempdata;
  float *tempvec, *tempconved;
  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {1/sqrt(2), -1/sqrt(2)};
  clock_t time_1, time_2;
  double cputime;


  tempdata = multialloc(sizeof(float), 2, band_size[0][0], band_size[0][1]);
  for ( i=0; i<band_size[0][0]; i=i+1 ) {
    for ( j=0; j<band_size[0][1]; j=j+1 ) {
      tempdata[i][j] = data[i][j];
    }
  }
  
  for ( n=0; n<levels; n=n+1 ) {
    time_1 = clock();
    l = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    h = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    for ( i=0; i<band_size[n][0]; i=i+1 ) {
      ext_conv_downsp(tempdata[i], band_size[n][1], lpfilter, 2, start, l[i]);
      ext_conv_downsp(tempdata[i], band_size[n][1], hpfilter, 2, start, h[i]);
    }
    time_2 = clock();
    cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
    multifree(tempdata, 2);
    tempdata = multialloc(sizeof(float), 2, band_size[n+1][0], band_size[n+1][1]);
    tempvec = malloc(sizeof(float)*band_size[n][0]);
    tempconved = malloc(sizeof(float)*band_size[n+1][0]);

    for ( j=0; j<band_size[n+1][1]; j=j+1 ) {
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = l[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        tempdata[i][j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][j] = tempconved[i];
      }
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = h[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[i][band_pos[n+1][1]+j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][band_pos[n+1][1]+j] = tempconved[i];
      }
    }
    multifree(h, 2);
    multifree(l, 2);
    free(tempvec);
    free(tempconved);
  }
  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      wave_coef[i][j] = tempdata[i][j];
    }
  }
  multifree(tempdata, 2);
}

      

/* this haar transform only handles image size of power of 2 */
void haar_transform(float **data, float *lpfilter, float *hpfilter, 
		    int filterlen_l, int filterlen_h, int lp_origin, 
		    int hp_origin, int height, int width, int levels, int start)
{
  int i, j, n;
  int subwidth, subheight;
  float **l;
  float **h;
  float **tempdata;
  float *conved;
  float *temp;

  for (n=0; n<levels; n=n+1 ) {
    subheight = height>>n;
    subwidth = width>>n;
    tempdata = multialloc(sizeof(float), 2, subheight, subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      for ( j=0; j<subwidth; j=j+1 ) {
	tempdata[i][j] = data[i][j];
      }
    }

    l = multialloc(sizeof(float), 2, subheight, subwidth/2);
    h = multialloc(sizeof(float), 2, subheight, subwidth/2);
    conved = malloc(sizeof(float)*subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], lpfilter, filterlen_l, lp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	l[i][j] = conved[start+j*2];
      }
    }
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], hpfilter, filterlen_h, hp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	h[i][j] = conved[start+j*2];
      }
    }
    free(conved);
    conved = malloc(sizeof(float)*subheight);
    temp = malloc(sizeof(float)*subheight);
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = l[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j] = conved[start+i*2];
      }
    }
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = h[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j+subwidth/2] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j+subwidth/2] = conved[start+i*2];
      }
    }
    free(conved);
    free(temp);
    multifree(tempdata, 2);
    multifree(l, 2);
    multifree(h, 2);
  }
}




/********* Main **********/
int main(int argc, char *argv[])
{
  FILE *fp, *fq;
  int i, j;
  float **image, **wave_coef;
  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {-1/sqrt(2), 1/sqrt(2)};
  int lp_origin;
  int hp_origin;
  int height, width, levels;
  float total_norm, x_org, y_org, x, y, x0, y0, pixel_size, energy;
  struct power_LUT *powertable;
  struct parameter para; 
  clock_t time_1, time_2;
  double cputime;
  int **band_size;
  int **band_pos;
  int height_conved, width_conved;
  char imgfile[50], outputfile[50];

  lp_origin = 0;
  hp_origin = 0;

  if (argc<1) {
    printf("need argument\n");
    return -1;
  }
  if ( (fp = fopen(argv[1], "r")) ) {
    fscanf(fp, "%f", &(para.a));
    fscanf(fp, "%f", &(para.b));
    fscanf(fp, "%f", &(para.f));
    fscanf(fp, "%f", &(para.g));
    fscanf(fp, "%d", &height);
    fscanf(fp, "%d", &width);
    fscanf(fp, "%d", &levels);
    fscanf(fp, "%s", imgfile);
    fscanf(fp, "%s", outputfile);
    fscanf(fp, "%f", &x);
    fscanf(fp, "%f", &y);
    fclose(fp);
  }
  else {
    printf("cannot open parameter file\n");
    exit(1);
  }
  para.a = para.a*1e-6;
  para.b = para.b*1e-6;
  pixel_size = IMG_height/(float)height;
  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;
  powertable = build_power_LUT(para.f, MAXVAL, SAMPLES);
  band_size = multialloc(sizeof(int), 2, levels+1, 2);
  band_pos = multialloc(sizeof(int), 2, levels+1, 2);
  image = multialloc(sizeof(float), 2, height, width);

  height_conved = 0;
  width_conved = 0;
  band_size[0][0] = height;
  band_size[0][1] = width;
  for ( i=1; i<=levels; i=i+1 ) {
    band_size[i][0] = floor((band_size[i-1][0]+1)/2);
    band_size[i][1] = floor((band_size[i-1][1]+1)/2);
  }
  band_pos[levels][0] = band_size[levels][0];
  band_pos[levels][1] = band_size[levels][1];
  for ( i=levels-1; i>=0; i=i-1 ) {
    band_pos[i][0] = band_pos[i+1][0]+band_size[i+1][0];
    band_pos[i][1] = band_pos[i+1][1]+band_size[i+1][1];
  }
  height_conved = band_pos[0][0];
  width_conved = band_pos[0][1];

  wave_coef = multialloc(sizeof(float), 2, height_conved, width_conved);

  if (PSF_IMG) {  
    fp = fopen("psffile.dat", "w");
    for ( i=0; i<height; i=i+1 ) {
      for ( j=0; j<width; j=j+1 ) {
        y0 = (float)i-y_org;
        x0 = (float)j-x_org;
        image[i][j] = computepsf(para, powertable, x0, y0, x, y, pixel_size);
        total_norm = total_norm + image[i][j];
        fprintf(fp, "%12.8f\t", image[i][j]);
      }
      fprintf(fp, "\n");
    }
    fclose(fp);
    printf("Sum of all values = %f\n", total_norm);
  }
  else {
    if ( (fp = fopen(imgfile, "r")) ) {
      for ( i=0; i<height; i=i+1 ) {
        for ( j=0; j<width; j=j+1 ) {
          fscanf(fp, "%f", &image[i][j]);
        }
      }
      fclose(fp);
    }
    else {
      printf("cannot open image file %s to read\n", imgfile);
      exit(1);
    }
  }

  energy = 0;
  for ( i=0; i<height; i=i+1 ) {
    for ( j=0; j<width; j=j+1 ) {
      energy += image[i][j]*image[i][j];
    }
  }
  printf("energy of input image = %f\n", energy);
  

  time_1 = clock();
  new_haar_transform(image, levels, 1, band_size, band_pos, wave_coef);
 
  /* haar_transform(image, lpfilter, hpfilter, 2, 2, lp_origin, 
		 hp_origin, height, width, levels, 1);
  */
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("cpu time for computing Haar transform with filter bank %f\n", cputime);

  fq = fopen(outputfile, "w");
  for ( i=0; i<height_conved; i=i+1 ) {
    for ( j=0; j<width_conved; j=j+1 ) {
      fprintf(fq, "%12.6f\t", wave_coef[i][j]);
    }
    fprintf(fq, "\n");
  }
  fclose(fq);

  energy = 0;
  for ( i=0; i<height_conved; i=i+1 ) {
    for ( j=0; j<width_conved; j=j+1 ) {
      energy += wave_coef[i][j]*wave_coef[i][j];
    }
  }
  printf("energy of transformed image = %f\n", energy); 

  multifree(image, 2);
  multifree(wave_coef, 2);
  multifree(band_size, 2);
  multifree(band_pos, 2);
  distroy_power_LUT(powertable);
  return 0;

}
